@@ -3,6 +3,7 @@
 module for mul and mulfix class: fund combination management
 """
 
+import logging
 import pandas as pd
 from pyecharts import options as opts
 from pyecharts.charts import Pie, ThemeRiver
@@ -12,8 +13,21 @@ from xalpha.evaluate import evaluate
 from xalpha.exceptions import FundTypeError, TradeBehaviorError
 from xalpha.record import record, irecord
 from xalpha.indicator import indicator
-from xalpha.info import cashinfo, fundinfo, mfundinfo
-from xalpha.trade import bottleneck, trade, turnoverrate, vtradevolume, xirrcal, itrade
+from xalpha.info import cashinfo, fundinfo, mfundinfo, get_fund_holdings
+from xalpha.trade import (
+    bottleneck,
+    trade,
+    turnoverrate,
+    vtradevolume,
+    xirrcal,
+    itrade,
+    vtradecost,
+)
+from xalpha.universal import get_fund_type, ttjjcode, get_rt
+import xalpha.universal as xu
+
+
+logger = logging.getLogger(__name__)
 
 
 class mul:
@@ -64,7 +78,9 @@ class mul:
             # unless you are sure corresponding funds are added to the droplist
         fundcodelist = [f.code for f in fundtradeobj]
         if status is not None:
-            for code in status.columns[1:]:
+            for code in status.columns:
+                if code == "date":
+                    continue
                 # r1, d2, v4 p = r+d+v
                 if code in fundcodelist:
                     continue
@@ -215,11 +231,14 @@ class mul:
         df = df.reset_index(drop=True)
         return df
 
-    def xirrrate(self, date=yesterdayobj(), guess=0.1):
+    def xirrrate(self, date=yesterdayobj(), startdate=None, guess=0.01):
         """
         xirr rate evauation of the whole invest combination
+
+        :param date: string or obj of datetime, the virtually sell-all date
+        :param startdate: string or obj of datetime, the beginning date of calculation, default from first buy
         """
-        return xirrcal(self.totcftable, self.fundtradeobj, date, guess)
+        return xirrcal(self.totcftable, self.fundtradeobj, date, startdate, guess)
 
     def evaluation(self, start=None):
         """
@@ -235,6 +254,117 @@ class mul:
         )
         return case
 
+    def get_stock_holdings(
+        self, year=None, season=None, date=yesterdayobj(), threhold=100
+    ):
+        """
+        获取整个基金组合的底层股票持仓总和和细节，组合穿透
+
+        :param year: 基于的基金季报年份
+        :param season: 基于的基金季报季度
+        :param date: 默认昨天
+        :param threhold: 默认100。小于100元的底层股票将不在最后的结果中展示
+        :return: pd.DataFrame column: name, code, value, ratio
+        """
+        d = {}
+        if year is None or season is None:
+            rd = convert_date(date) - pd.Timedelta(days=120)
+            if not year:
+                year = rd.year
+            if not season:
+                season = int((rd.month - 0.1) / 3) + 1
+            logger.debug("use %s, %s for fund report" % (year, season))
+        for f in self.fundtradeobj:
+            if isinstance(f, itrade):
+                if f.get_type() == "股票":
+                    code = f.code
+                elif f.get_type() == "场内基金":
+                    code = f.code[2:]
+                else:
+                    continue
+            else:
+                code = f.code
+            value = f.briefdailyreport(date).get("currentvalue", 0)
+            if value > 0:
+                if code.startswith("SH") or code.startswith("SZ"):
+                    stock = code
+                    d[stock] = d.get(stock, 0) + value
+                elif code == "mf":
+                    continue
+                else:
+                    df = get_fund_holdings(code, year, season)
+                    if df is None:
+                        continue
+
+                    for _, row in df.iterrows():
+                        stock = row["code"]
+                        stock = ttjjcode(stock)
+                        d[stock] = d.get(stock, 0) + row["ratio"] / 100 * value
+                        # print("%s has %s contribution from %s" %(stock, row["ratio"] / 100 * value, f.name))
+
+        l = []
+        for code, value in sorted(d.items(), key=lambda item: -item[1]):
+            if value >= threhold:
+                try:
+                    name = get_rt(code)["name"]
+                except:
+                    name = code
+                l.append([name, code, value])
+        fdf = pd.DataFrame(l, columns=["name", "code", "value"])
+        fdf["ratio"] = fdf["value"] / fdf["value"].sum()
+        return fdf
+
+    def get_portfolio(self, date=yesterdayobj()):
+        """
+        获取基金组合底层资产大类配置的具体值
+
+        :param date:
+        :return: Dict[str, float]. stock，bond，cash 对应总值的字典
+        """
+
+        d = {"stock": 0, "bond": 0, "cash": 0}
+        date = convert_date(date)
+        for f in self.fundtradeobj:
+            value = f.briefdailyreport(date).get("currentvalue", 0)
+            if value > 0:
+                if isinstance(f, itrade):
+                    if f.get_type() == "股票":
+                        d["stock"] += value
+                        continue
+                    elif f.get_type() in ["可转债", "债券"]:
+                        d["bond"] += value
+                        continue
+                    elif f.get_type() == "货币基金":
+                        d["cash"] += value
+                        continue
+                    elif f.get_type() == "场内基金":
+                        code = f.code[2:]
+                    else:
+                        continue
+                else:
+                    code = f.code
+                if code == "mf":
+                    d["cash"] += value
+                    continue
+                if get_fund_type(code) == "货币基金":
+                    d["cash"] += value
+                    continue
+                df = xu.get_daily("pt-F" + code, end=date.strftime("%Y%m%d"))
+                if df is None or len(df) == 0:
+                    logger.warning("empty portfolio info for %s" % code)
+                row = df.iloc[-1]
+                if row["bond_ratio"] + row["stock_ratio"] < 10:  # 联接基金
+                    d["stock"] += (
+                        (100 - row["bond_ratio"] - row["cash_ratio"]) * value / 100
+                    )
+                    d["bond"] += row["bond_ratio"] * value / 100
+                    d["cash"] += row["cash_ratio"] * value / 100
+                else:
+                    d["stock"] += row["stock_ratio"] * value / 100
+                    d["bond"] += row["bond_ratio"] * value / 100
+                    d["cash"] += row["cash_ratio"] * value / 100
+        return d
+
     def v_positions(self, date=yesterdayobj(), rendered=True):
         """
         pie chart visualization of positions ratio in combination
@@ -267,6 +397,51 @@ class mul:
         else:
             return pie
 
+    def v_category_positions(self, date=yesterdayobj(), rendered=True):
+        """
+        资产分类扇形图，按大类资产求和绘制
+
+        :param date:
+        :param rendered: bool. default true for notebook, for plain pyechart obj to return, set rendered=False
+        :return:
+        """
+        d = {}
+        for f in self.fundtradeobj:
+            if isinstance(f, itrade):
+                t = f.get_type()
+                if t == "场内基金":
+                    t = get_fund_type(f.code[2:])
+            elif f.code == "mf":
+                t = "货币基金"
+            else:
+                t = get_fund_type(f.code)
+            if t == "其他":
+                logger.warning(
+                    "%s has category others which should be double checked" % f.code
+                )
+            d[t] = d.get(t, 0) + f.briefdailyreport(date).get("currentvalue", 0)
+
+        sdata = sorted([(k, round(v, 2)) for k, v in d.items()])
+        pie = Pie()
+        pie.add(
+            series_name="总值占比",
+            data_pair=sdata,
+            label_opts=opts.LabelOpts(is_show=False, position="center"),
+        ).set_global_opts(
+            legend_opts=opts.LegendOpts(
+                pos_left="left", type_="scroll", orient="vertical"
+            )
+        ).set_series_opts(
+            tooltip_opts=opts.TooltipOpts(
+                trigger="item", formatter="{a} <br/>{b}: {c} ({d}%)"
+            ),
+        )
+
+        if rendered:
+            return pie.render_notebook()
+        else:
+            return pie
+
     def v_positions_history(self, end=yesterdaydash(), rendered=True):
         """
         river chart visulization of positions ratio history
@@ -399,6 +574,14 @@ class mulfix(mul, indicator):
             res += fund.briefdailyreport(date).get("currentvalue", 0)
         return res / self.totmoney
 
+    def v_tradecost(self, threhold=0, date=yesterdayobj(), rendered=True):
+        if getattr(self, "price", None) is None:
+            raise ValueError("Please generate price table by ``bcmkset()`` first")
+        cftable = self.fundtradeobj[-1].cftable[1:]
+        cftable = cftable[abs(cftable["cash"]) > threhold]
+        cftable["cash"] = -cftable["cash"]
+        return vtradecost(self, cftable, end=date, rendered=rendered)
+
 
 class imul(mul):
     def __init__(self, *fundtradeobj, status=None, istatus=None):
