@@ -3,6 +3,7 @@
 modules for Object oriented toolbox which wrappers get_daily and some more
 """
 
+import re
 import sys
 import datetime as dt
 import numpy as np
@@ -11,6 +12,7 @@ from collections import deque
 from functools import wraps, lru_cache
 import logging
 from scipy import stats
+from bs4 import BeautifulSoup
 
 from xalpha.cons import (
     opendate,
@@ -20,12 +22,17 @@ from xalpha.cons import (
     scale_dict,
     tz_bj,
     holidays,
+    xnpv,
+    xirr,
+    rget,
+    rpost,
 )
 from xalpha.info import get_fund_holdings
 from xalpha.universal import (
     get_rt,
     get_bar,
     ttjjcode,
+    get_bond_rates,
     _convert_code,
     _inverse_convert_code,
     fetch_backend,
@@ -167,6 +174,7 @@ class IndexPEBHistory:
         "000300.XSHG": ("沪深300", "2012-01-01"),
         "000905.XSHG": ("中证500", "2012-01-01"),
         "000922.XSHG": ("中证红利", "2012-01-01"),
+        "000925.XSHG": ("基本面50", "2012-01-01"),
         "399006.XSHE": ("创业板指", "2012-01-01"),
         "000992.XSHG": ("全指金融", "2012-01-01"),
         "000991.XSHG": ("全指医药", "2012-01-01"),
@@ -613,6 +621,295 @@ class OverPriced:
         return ax
 
 
+#########################
+# cb value estimation   #
+#########################
+
+
+def BlackScholes(S, K, t, v, r=0.02, CallPutFlag="C"):
+    """
+    BS option pricing calculator
+
+    :param S: current stock price
+    :param K: stricking price
+    :param t: Time until option exercise (years to maturity)
+    :param r: risk-free interest rate (by year)
+    :param v: Variance(volitility) of annual increase
+    :param CallPutFlag: "C" or "P", default call option
+    :return:
+    """
+    # function modified from https://github.com/boyac/pyOptionPricing
+
+    def CND(X):
+        return stats.norm.cdf(X)
+
+    d1 = (np.log(S / K) + (r + (v ** 2) / 2) * t) / (v * np.sqrt(t))
+    d2 = d1 - v * np.sqrt(t)
+
+    if CallPutFlag in ["c", "C"]:
+        return S * CND(d1) - K * np.exp(-r * t) * CND(d2)  # call option
+    elif CallPutFlag in ["p", "P"]:
+        return K * np.exp(-r * t) * CND(-d2) - S * CND(-d1)  # put option
+    else:
+        raise ValueError("Unknown CallPutFlag %s" % CallPutFlag)
+
+
+def cb_bond_value(issue_date, rlist, rate=0.03, date=None, tax=1.0):
+    """
+    可转债债券价值计算器
+
+    :param issue_date: str. 发行日期
+    :param rlist: List[float], 每年度的利息百分点，比如 0.4,0.6等，最后加上最后返回的值（不含最后一年利息），比如 104
+    :param rate:  float，现金流折算利率，应取同久期同信用等级的企业债利率，参考 https://yield.chinabond.com.cn/
+    :param date: 默认今天，计算债券价值基于的时间
+    :param tax: float，税率，1.0 表示不算税后，0.8 为计算税后利息，一般不需要设置成0.8，因为区别不大
+    :return:
+    """
+
+    if rlist[-1] < 100:
+        logger.warning(
+            "the format of rlist must contain the final return more than 100 without interest of that year"
+        )
+    issue_date = issue_date.replace("-", "").replace("/", "")
+    issue_date_obj = dt.datetime.strptime(issue_date, "%Y%m%d")
+    if date is None:
+        date_obj = dt.datetime.today()
+    else:
+        date = date.replace("-", "").replace("/", "")
+        date_obj = dt.datetime.strptime(date, "%Y%m%d")
+    cf = [(date_obj, 0)]
+    passed = (date_obj - issue_date_obj).days // 365
+    for i, r in enumerate(rlist[:-1]):
+        if i >= passed:
+            cf.append((issue_date_obj + dt.timedelta(days=(i + 1) * 365), r * tax))
+    cf.append((issue_date_obj + dt.timedelta(days=(len(rlist) - 1) * 365), rlist[-1]))
+    return xnpv(rate, cf)
+
+
+def cb_ytm(issue_date, rlist, cp, date=None, tax=1.0, guess=0.01):
+    """
+    可转债到期收益率计算器
+
+    :param issue_date: 发行日期
+    :param rlist: 计息及赎回列表
+    :param cp: 可转债现价
+    :param date: 参考日期
+    :param tax: 计税 1 vs 0.8 税后 YTM
+    :param guess: YTM 估计初始值
+    :return:
+    """
+
+    if rlist[-1] < 100:
+        logger.warning(
+            "the format of rlist must contain the final return more than 100 without interest of that year"
+        )
+    issue_date = issue_date.replace("-", "").replace("/", "")
+    issue_date_obj = dt.datetime.strptime(issue_date, "%Y%m%d")
+    if date is None:
+        date_obj = dt.datetime.today()
+    else:
+        date = date.replace("-", "").replace("/", "")
+        date_obj = dt.datetime.strptime(date, "%Y%m%d")
+    cf = [(date_obj, -cp)]
+    passed = (date_obj - issue_date_obj).days // 365
+    for i, r in enumerate(rlist[:-1]):
+        if i >= passed:
+            cf.append((issue_date_obj + dt.timedelta(days=(i + 1) * 365), r * tax))
+    # 关于赎回利息计算： https://www.jisilu.cn/?/question/339
+    # https://www.jisilu.cn/question/5807
+    # 富投网的算法：将最后一年超出100的部分，全部按照20%计税，
+    cf.append((issue_date_obj + dt.timedelta(days=(len(rlist) - 1) * 365), rlist[-1]))
+    #     print(cf)
+    try:
+        return xirr(cf, guess=guess)
+    except RuntimeError:
+        return
+
+
+class CBCalculator:
+    """
+    可转债内在价值，简单计算器，期权价值与债券价值估算
+    """
+
+    def __init__(
+        self,
+        code,
+        bondrate=None,
+        riskfreerate=None,
+        volatility=None,
+        name=None,
+        zgj=None,
+    ):
+        """
+
+        :param code: str. 转债代码，包含 SH 或 SZ 字头
+        :param bondrate: Optional[float]. 评估所用的债券折现率，默认使用中证企业债对应信用评级对应久期的利率
+        :param riskfreerate: Optioal[float]. 评估期权价值所用的无风险利率，默认使用国债对应久期的年利率。
+        :param volatility: Optional[float]. 正股波动性百分点，默认在一个范围浮动加上历史波动率的小幅修正。
+        :param name: str. 对于历史回测，可以直接提供 str，免得多次 get_rt 获取 name
+        :param zgj: float. 手动设置转股价，适用于想要考虑转股价调整因素进行历史估值的高阶用户
+        """
+        # 应该注意到该模型除了当天外，其他时间估计会利用现在的转股价，对于以前下修过转股价的转债历史价值估计有问题
+
+        self.code = code
+        self.refbondrate = bondrate
+        self.bondrate = self.refbondrate
+        self.refriskfreerate = riskfreerate
+        self.riskfreerate = self.refriskfreerate
+        self.refvolatility = volatility
+        self.volatility = self.refvolatility
+        self.name = name
+
+        r = rget("https://www.jisilu.cn/data/convert_bond_detail/" + code[2:])
+        r.encoding = "utf-8"
+        b = BeautifulSoup(r.text, "lxml")
+        self.rlist = [
+            float(re.search(r"[\D]*([\d]*.[\d]*)[\s]*\%", s).group(1))
+            for s in re.split("、|，", b.select("td[id=cpn_desc]")[0].string)
+        ]
+        self.rlist.append(float(b.select("td[id=redeem_price]")[0].string))
+        self.rlist[-1] -= self.rlist[-2]  # 最后一年不含息返多少
+        self.scode = (
+            b.select("td[class=jisilu_nav]")[0].contents[1].string.split("-")[1].strip()
+        )
+        self.scode = ttjjcode(self.scode)  # 标准化股票代码
+        if not zgj:
+            self.zgj = float(b.select("td[id=convert_price]")[0].string)  # 转股价
+        else:
+            self.zgj = zgj
+        self.rating = b.select("td[id=rating_cd]")[0].string.strip()
+        self.enddate = b.select("td[id=maturity_dt]")[0].string
+
+    def process_byday(self, date=None):
+        if not date:
+            self.date_obj = dt.datetime.today()
+        else:
+            self.date_obj = dt.datetime.strptime(
+                date.replace("-", "").replace("/", ""), "%Y%m%d"
+            )
+        if not date:
+            rt = get_rt(self.code)
+            self.name = rt["name"]
+            self.cbp = rt["current"]  # 转债价
+            self.stockp = get_rt(self.scode)["current"]  # 股票价
+        else:
+            try:
+                if not self.name:
+                    rt = get_rt(self.code)
+                    self.name = rt["name"]
+            except:
+                self.name = "unknown"
+            df = xu.get_daily(self.code, prev=100, end=self.date_obj.strftime("%Y%m%d"))
+            self.cbp = df.iloc[-1]["close"]
+            df = xu.get_daily(
+                self.scode, prev=100, end=self.date_obj.strftime("%Y%m%d")
+            )
+            self.stockp = df.iloc[-1]["close"]
+
+        df = xu.get_daily(self.scode, prev=360, end=self.date_obj.strftime("%Y%m%d"))
+        self.history_volatility = np.std(
+            np.log(df["close"] / df.shift(1)["close"])
+        ) * np.sqrt(244)
+        if not self.refvolatility:
+            self.volatility = 0.17
+            if self.rating in ["A-", "A", "A+"]:
+                self.volatility = 0.25
+            elif self.rating in ["AA-"]:
+                self.volatility = 0.2
+            elif self.rating in ["AA"]:
+                self.volatility = 0.19
+            elif self.rating in ["AA+"]:
+                self.volatility = 0.18
+            if self.history_volatility < 0.2:
+                self.volatility -= 0.01
+            elif self.history_volatility > 0.7:
+                self.volatility += 0.05
+            elif self.history_volatility > 0.6:
+                self.volatility += 0.035
+            elif self.history_volatility > 0.5:
+                self.volatility += 0.02
+            elif self.history_volatility > 0.4:
+                self.volatility += 0.01
+        self.years = len(self.rlist) - 1
+        syear = int(self.enddate.split("-")[0]) - self.years
+        self.issuedate = str(syear) + self.enddate[4:]
+        self.days = (
+            dt.datetime.strptime(self.enddate, "%Y-%m-%d") - self.date_obj
+        ).days
+        if not self.refbondrate:
+            ratestable = get_bond_rates(self.rating, self.date_obj.strftime("%Y-%m-%d"))
+            if self.rating in ["A", "A+", "AA-"]:
+                ## AA 到 AA- 似乎是利率跳高的一个坎
+                cutoff = 2
+            else:
+                cutoff = 4
+            if self.days / 365 > cutoff:
+                # 过长久期的到期收益率，容易造成估值偏离，虽然理论上是对的
+                # 考虑到国内可转债市场信用风险较低，不应过分低估低信用债的债券价值
+                self.bondrate = (
+                    ratestable[ratestable["year"] <= cutoff].iloc[-1]["rate"] / 100
+                )
+            else:
+                self.bondrate = (
+                    ratestable[ratestable["year"] >= self.days / 365].iloc[0]["rate"]
+                    / 100
+                )
+        if not self.refriskfreerate:
+            ratestable = get_bond_rates("N", self.date_obj.strftime("%Y-%m-%d"))
+            if self.days / 365 > 5:
+                self.riskfreerate = (
+                    ratestable[ratestable["year"] <= 5].iloc[-1]["rate"] / 100
+                )
+            else:
+                self.riskfreerate = (
+                    ratestable[ratestable["year"] >= self.days / 365].iloc[0]["rate"]
+                    / 100
+                )
+
+    def analyse(self, date=None):
+        self.process_byday(date=date)
+        d = {
+            "stockcode": self.scode,
+            "cbcode": self.code,
+            "name": self.name,
+            "enddate": self.enddate,
+            "interest": self.rlist,
+            "zgj": self.zgj,
+            "stockprice": self.stockp,
+            "cbprice": self.cbp,
+            "rating": self.rating,
+            "bondrate": self.bondrate,
+            "predicted_volatility": self.volatility,
+            "historical_volatility": self.history_volatility,
+            "riskfreerate": self.riskfreerate,
+            "years": self.days / 365,
+            "issuedate": self.issuedate,
+            "date": self.date_obj.strftime("%Y-%m-%d"),
+        }
+        d["bond_value"] = cb_bond_value(self.issuedate, self.rlist, self.bondrate)
+        d["ytm_wo_tax"] = cb_ytm(self.issuedate, self.rlist, self.cbp)
+        d["ytm_wi_tax"] = cb_ytm(self.issuedate, self.rlist, self.cbp, tax=0.8)
+        d["option_value"] = (
+            BlackScholes(
+                self.stockp,
+                self.zgj,
+                self.days / 365,
+                self.volatility,
+                self.riskfreerate,
+                CallPutFlag="C",
+            )
+            * 100
+            / self.zgj
+        )
+        # 经验上看，下修强赎回售及美式期权行为等其他带来的期权价值大约有1到4元的增益：
+        # 以0.015 为无风险利率和 0.15-0.18 为波动率估计范围的情形下
+        # 实在没有必要为了这几块钱上复杂工具估值，因为无风险利率几十个bp的改变，就足以导致更大的波动，看个热闹就行了
+        # 可转债估值只能是模糊的正确
+        d["tot_value"] = d["bond_value"] + d["option_value"]
+        d["premium"] = (self.cbp / d["tot_value"] - 1) * 100
+        return d
+
+
 #########################
 # netvalue prediction   #
 #########################
@@ -858,8 +1155,11 @@ def daily_increment(code, date, lastday=None, _check=False, warning_threhold=Non
         ratio = tds.iloc[-1]["close"] / tds.iloc[-2]["close"]
     else:
         tds2 = tds[tds["date"] <= lastday]
+        if tds2 is None or len(tds2) == 0:
+            ratio = 1.0
         # 未考虑检查连 lastday 的数据数据源都没更新的情形，这种可能极小
-        ratio = tds.iloc[-1]["close"] / tds2.iloc[-1]["close"]
+        else:
+            ratio = tds.iloc[-1]["close"] / tds2.iloc[-1]["close"]
     return ratio
 
 
@@ -1022,13 +1322,14 @@ class RTPredict:
         else:
             return self.t1value_cache[0]
 
+    @error_catcher
     def get_t0(self, return_date=True, percent=False):
         last_value, last_date = self.get_t1()
         last_date_obj = dt.datetime.strptime(last_date, "%Y-%m-%d")
         cday = last_onday(self.today)
         while last_date_obj < cday:  # 昨天净值数据还没更新
             # 是否存在部分部分基金可能有 gap？
-            if cday.strftime("%Y-%m-%d") not in gap_info[self.fcode]:
+            if cday.strftime("%Y-%m-%d") not in gap_info.get(self.fcode, []):
                 self.t1_type = "昨日未出"
                 raise DateMismatch(
                     self.code,
